#![warn (clippy::pedantic)]

use std::{
	collections::*,
	ffi::OsString,
	net::SocketAddr,
	path::{PathBuf},
	sync::Arc,
};

use arc_swap::ArcSwap;
use hyper::{
	Body,
	Request,
	Response,
	Server,
	service::{
		make_service_fn,
		service_fn,
	},
	StatusCode,
};
use serde::Deserialize;
use tokio_stream::wrappers::ReceiverStream;
use tracing::debug;

use ptth_core::{
	http_serde::RequestParts,
	prelude::*,
};
use ptth_server::{
	file_server::{
		self,
		metrics,
		FileServer,
	},
	load_toml,
};

async fn handle_all (req: Request <Body>, state: Arc <FileServer>) 
-> anyhow::Result <Response <Body>>
{
	use std::str::FromStr;
	use hyper::header::HeaderName;
	
	debug! ("req.uri () = {:?}", req.uri ());
	
	let path_and_query = req.uri ().path_and_query ().map_or_else (|| req.uri ().path (), http::uri::PathAndQuery::as_str);
	
	let path_and_query = path_and_query.into ();
	
	let (parts, _) = req.into_parts ();
	
	let ptth_req = RequestParts::from_hyper (parts.method, path_and_query, parts.headers)?;
	
	let ptth_resp = state.serve_all (
		ptth_req.method, 
		&ptth_req.uri, 
		&ptth_req.headers
	).await?;
	
	let mut resp = Response::builder ()
	.status (StatusCode::from (ptth_resp.parts.status_code));
	
	for (k, v) in ptth_resp.parts.headers {
		resp = resp.header (HeaderName::from_str (&k)?, v);
	}
	
	let body = ptth_resp.body.map_or_else (Body::empty, |body| {
		Body::wrap_stream (ReceiverStream::new (body))
	});
	
	Ok (resp.body (body)?)
}

#[derive (Deserialize)]
struct ConfigFile {
	file_server_root: Option <PathBuf>,
	file_server_roots: Option <BTreeMap <String, PathBuf>>,
	name: Option <String>,
}

pub async fn main (_args: &[OsString]) -> anyhow::Result <()> {
	let path = PathBuf::from ("./config/ptth_server.toml");
	
	let file_server_root;
	let file_server_roots;
	let name;
	
	match load_toml::load::<ConfigFile, _> (&path) {
		Ok (config_file) => {
			file_server_root = config_file.file_server_root;
			file_server_roots = config_file.file_server_roots;
			name = config_file.name;
		},
		_ => {
			info! ("No ptth_server.toml file, using default configs");
			file_server_root = None;
			file_server_roots = None;
			name = None;
		},
	};
	
	let file_server_root = file_server_root.unwrap_or_else (|| PathBuf::from ("."));
	let file_server_roots = file_server_roots.unwrap_or_else (|| Default::default ());
	let name = name.unwrap_or_else (|| "PTTH File Server".to_string ());
	
	info! ("file_server_root: {:?}", file_server_root);
	
	let addr = SocketAddr::from(([0, 0, 0, 0], 4000));
	info! ("Serving at {:?}", addr);
	
	let metrics_interval = Arc::new (ArcSwap::default ());
	
	let interval_writer = Arc::clone (&metrics_interval);
	tokio::spawn (async move {
		file_server::metrics::Interval::monitor (interval_writer).await;
	});
	
	let config = file_server::Config {
		file_server_root,
		file_server_roots,
	};
	
	let state = Arc::new (FileServer::new (
		config,
		&PathBuf::new (),
		name,
		metrics_interval,
		Some (path),
	)?);
	
	let make_svc = make_service_fn (|_conn| {
		let state = state.clone ();
		
		async {
			Ok::<_, String> (service_fn (move |req| {
				let state = state.clone ();
				
				handle_all (req, state)
			}))
		}
	});
	
	let (shutdown_rx, forced_shutdown) = ptth_core::graceful_shutdown::init_with_force ();
	
	let server = Server::bind (&addr)
	.serve (make_svc)
	.with_graceful_shutdown (async move {
		shutdown_rx.await.ok ();
	});
	
	forced_shutdown.wrap_server (server).await??;
	
	Ok (())
}

